import base64
import json
import logging
import os
import time
from pathlib import Path
from typing import Mapping, Optional, Union, cast

import requests
from httpx import URL, Client, Limits, Timeout
from tqdm import tqdm

from ._base_client import BaseClient, SyncHttpxClientWrapper
from ._constants import (
    DEFAULT_CONNECTION_LIMITS,
    DEFAULT_MAX_RETRIES,
    DEFAULT_TIEFBLUE_TIMEOUT,
)
from ._response import APIResponse

_DEFAULT_CHUNK_SIZE = 50 * 1024 * 1024
_DEFAULT_ITERATE_MAX_OBJECTS = 50

logger = logging.getLogger(__name__)


class Parameter(object):
    contentType: str
    contentEncoding: str
    contentLanguage: str
    contentDisposition: str
    cacheControl: str
    acl: str
    expires: str
    userMeta: dict
    predefinedMeta: str


class Tiefblue(BaseClient[Client]):
    TIEFBLUE_HEADER_KEY = "X-Storage-Param"
    _client: Client

    def __init__(
        self,
        base_url: Optional[Union[str, URL]] = None,
        max_retries: int = DEFAULT_MAX_RETRIES,
        timeout: Union[float, Timeout, None] = DEFAULT_TIEFBLUE_TIMEOUT,
        limits: Limits = DEFAULT_CONNECTION_LIMITS,
        _version: Optional[str] = None,
        custom_headers: Optional[Mapping[str, str]] = None,
    ) -> None:

        if base_url is None:
            base_url = os.environ.get("TIEFBLUE_BASE_URL")

        if base_url is None:
            base_url = "https://tiefblue.dp.tech"

        super().__init__(
            _version=_version,
            limits=limits,
            # cast to a valid type because mypy doesn't understand our type narrowing
            timeout=cast(Timeout, timeout),
            base_url=base_url,
            max_retries=max_retries,
            custom_headers=custom_headers,
        )
        self._client = SyncHttpxClientWrapper(
            base_url=base_url,
            # cast to a valid type because mypy doesn't understand our type narrowing
            timeout=cast(Timeout, timeout),
            limits=limits,
            follow_redirects=True,
        )

    def encode_base64(self, parameter: dict = {}) -> str:
        j = json.dumps(parameter)
        return base64.b64encode(j.encode()).decode()

    def _write(
        self,
        object_key: str,
        data: str = "",
        headers: dict = {},
        parameter: dict = {},
    ):
        param = {"path": object_key}

        if parameter:
            param["option"] = parameter.__dict__

        headers[self.TIEFBLUE_HEADER_KEY] = self.encode_base64(param)
        resp = self._client.post("/api/upload/binary", headers=headers, data=data)

        return APIResponse(resp).json

    def read(self, object_key: str = "", token: str = "", ranges: str = "") -> None:

        url = f"/api/download/{object_key}"
        self.client.token = token
        res = self.client.get(url, host=self.host, stream=True)
        return res

    def upload_from_file(
        self,
        object_key: str = "",
        file_path: str = "",
        token: str = "",
        parameter: dict = None,
    ) -> None:
        if not os.path.exists(file_path):
            raise FileNotFoundError
        if os.path.isdir(file_path):
            raise IsADirectoryError
        _, disposition = os.path.split(file_path)
        if parameter is None:
            parameter = Parameter()
        parameter.contentDisposition = f'attachment; filename="{disposition}"'
        with open(file_path, "rb") as fp:
            res = self._write(
                object_key=object_key, data=fp.read(), parameter=parameter
            )
            return res

    def init_upload_by_part(self, object_key: str, parameter=None):
        data = {"path": object_key}
        if parameter is not None:
            data["option"] = parameter.__dict__
        url = "/api/upload/multipart/init"
        return self._client.post(url, host=self.host, json=data)

    def upload_by_part(
        self, object_key: str, initial_key: str, chunk_size: int, number: int, body
    ):
        param = {
            "initialKey": initial_key,
            "number": number,
            "partSize": chunk_size,
            "objectKey": object_key,
        }
        headers = {}
        headers[self.TIEFBLUE_HEADER_KEY] = self._dump_parameter(param)
        url = "/api/upload/multipart/upload"
        resp = self._client.post(url, host=self.host, data=body, headers=headers)
        return resp

    def complete_upload_by_part(self, object_key, initial_key, part_string):
        data = {
            "path": object_key,
            "initialKey": initial_key,
            "partString": part_string,
        }
        url = "/api/upload/multipart/complete"
        resp = self.client.post(url, host=self.host, json=data)
        return resp

    def upload_from_file_multi_part(
        self,
        *,
        object_key: Union[str, Path],
        file_path: Union[str, Path],
        custom_headers: Optional[Mapping[str, str]] = None,
        chunk_size: int = _DEFAULT_CHUNK_SIZE,
        parameter=None,
        progress_bar=False,
        need_parse=False,
    ) -> None:
        if isinstance(file_path, Path):
            if not file_path.exists():
                raise FileNotFoundError
        elif not os.path.exists(file_path):
            raise FileNotFoundError
        if os.path.isdir(file_path):
            raise IsADirectoryError
        object_key = str(object_key)
        if need_parse:
            _, _, object_key = self._parse_ap_name_and_tag(object_key)
        size = os.path.getsize(file_path)
        _, disposition = os.path.split(file_path)
        if parameter is None:
            parameter = Parameter()
        parameter.contentDisposition = f'attachment; filename="{disposition}"'
        bar_format = "{l_bar}{bar}| {n:.02f}/{total:.02f} %  [{elapsed}<{remaining}, {rate_fmt}{postfix}]"
        with open(file_path, "r") as f:
            pbar = tqdm(
                total=100,
                desc=f"Uploading {disposition}",
                smoothing=0.01,
                bar_format=bar_format,
                disable=not progress_bar,
            )
            f.seek(0)
            if size < _DEFAULT_CHUNK_SIZE * 2:
                self._write(
                    object_key=object_key,
                    data=f.buffer,
                    parameter=parameter,
                    headers=custom_headers,
                )
                pbar.update(100)
                pbar.close()
                return
            chunks = split_size_by_part_size(size, chunk_size)
            initial_key = self.init_upload_by_part(object_key, parameter).get(
                "initialKey"
            )
            part_string = []
            for c in chunks:
                f.seek(c.Offset)
                num_to_upload = min(chunk_size, size - c.Offset)
                part_string.append(
                    self.upload_by_part(
                        object_key,
                        initial_key,
                        chunk_size=c.Size,
                        number=c.Number,
                        body=f.buffer.read(c.Size),
                    ).get("partString")
                )
                percent = num_to_upload * 100 / (size + 1)
                pbar.update(percent)
            pbar.close()
            return self.complete_upload_by_part(object_key, initial_key, part_string)

    def download_from_file(self):
        pass

    def download_from_url(self, url: str, save_file: Path):
        ret = None
        for retry_count in range(3):
            try:
                ret = requests.get(url, stream=True)
            except Exception as e:
                logger.warning(f"download from url failed: {e}")
                continue
            if ret.ok:
                break
            else:
                time.sleep(retry_count)
                ret = None
        if ret is not None:
            ret.raise_for_status()
            with open(save_file, "wb") as f:
                for chunk in ret.iter_content(chunk_size=8192):
                    f.write(chunk)
            ret.close()

    def _dump_parameter(self, parameter):
        j = json.dumps(parameter)
        return base64.b64encode(j.encode()).decode()

    def decode_base64(self, encode_data):
        data = json.loads(base64.b64decode(encode_data).decode("utf-8"))
        return data

    def iterate(self, header: dict = {}, body: dict = {}):
        uri = "/api/iterate"
        response = self.client.post(uri, headers=header, json=body)
        return APIResponse(response).json

    def _parse_ap_name_and_tag(self, input_path: str):
        input_path_arr = input_path.split("/")
        if len(input_path_arr) < 3:
            return "", "".input_path_arr
        return input_path_arr[0], input_path_arr[1], "/".join(input_path_arr[2:])


class Chunk:
    Number: int
    Offset: int
    Size: int


def split_size_by_part_size(total_size: int, chunk_size: int):
    if chunk_size < _DEFAULT_CHUNK_SIZE:
        chunk_size = _DEFAULT_CHUNK_SIZE
    chunk_number = int(total_size / chunk_size)
    if chunk_number >= 10000:
        raise TooManyChunk
    chunks = []
    for i in range(chunk_number):
        c = Chunk()
        c.Number = i + 1
        c.Offset = i * chunk_size
        c.Size = chunk_size
        chunks.append(c)

    if total_size % chunk_size > 0:
        c = Chunk()
        c.Number = len(chunks) + 1
        c.Offset = len(chunks) * chunk_size
        c.Size = total_size % chunk_size
        chunks.append(c)
    return chunks


def partial_with_start_from(start_bytes):
    return f"bytes={start_bytes}-"


def partial_with_end_from(end_bytes):
    return f"bytes=-{end_bytes}"


def partial_with_range(start_bytes, end_bytes):
    return f"bytes={start_bytes}-{end_bytes}"


TooManyChunk = Exception("too many chunks, please consider increase your chunk size")



import logging
import platform
import time
from functools import wraps
import traceback
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    Literal,
    Mapping,
    Optional,
    TypeVar,
    Union,
    cast,
    Callable,
)
from urllib.parse import urljoin

import distro
import httpx
from httpx import URL, Limits, Timeout


from ._constants import DEFAULT_CONNECTION_LIMITS, DEFAULT_MAX_RETRIES, DEFAULT_TIMEOUT

logger = logging.getLogger(__name__)

Arch = Union[Literal["x32", "x64", "arm", "arm64", "unknown"]]

Platform = Union[
    Literal[
        "MacOS",
        "Linux",
        "Windows",
        "FreeBSD",
        "OpenBSD",
        "iOS",
        "Android",
        "Unknown",
    ],
]


class _DefaultHttpxClient(httpx.Client):
    def __init__(self, **kwargs: Any) -> None:
        kwargs.setdefault("timeout", DEFAULT_TIMEOUT)
        kwargs.setdefault("limits", DEFAULT_CONNECTION_LIMITS)
        kwargs.setdefault("follow_redirects", True)
        super().__init__(**kwargs)


if TYPE_CHECKING:
    DefaultHttpxClient = httpx.Client
    """An alias to `httpx.Client` that provides the same defaults that this SDK
    uses internally.

    This is useful because overriding the `http_client` with your own instance of
    `httpx.Client` will result in httpx's defaults being used, not ours.
    """
else:
    DefaultHttpxClient = _DefaultHttpxClient

_HttpxClientT = TypeVar("_HttpxClientT", bound=Union[httpx.Client, httpx.AsyncClient])


class SyncHttpxClientWrapper(DefaultHttpxClient):
    def __del__(self) -> None:
        try:
            self.close()
        except Exception:
            pass


def retry(max_retries_func, delay=1, exceptions=(Exception,)):
    """
    Decorator for retrying a function upon exception.

    max_retries: The maximum number of retries before giving up.
    delay: Initial delay between retries in seconds.
    backoff: Multiplier applied to delay between retries (e.g., for exponential backoff).
    exceptions: Tuple of exceptions to catch. Defaults to base Exception class.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            self = args[0]
            max_retries = max_retries_func(self)
            retries = 0
            while retries < max_retries:
                try:
                    return func(*args, **kwargs)
                except exceptions as e:
                    if retries >= max_retries:
                        logger.error(
                            f"Exhausted all retries. Raising the last caught exception: {e}"
                        )
                        raise e
                    else:
                        logger.warning(
                            f"Retry {retries} due to error: {e}. Retrying in {delay} seconds..."
                        )
                        time.sleep(delay)
                    retries += 1

        return wrapper

    return decorator


class BaseClient(Generic[_HttpxClientT]):
    _client: _HttpxClientT
    _base_url: URL
    max_retries: int
    timeout: Union[float, Timeout, None]
    _limits: httpx.Limits
    _version: Union[str, None]

    def __init__(
        self,
        *,
        _version: Optional[str] = None,
        base_url: Union[str, URL],
        limits: httpx.Limits,
        max_retries: int = DEFAULT_MAX_RETRIES,
        timeout: Optional[Union[float, httpx.Timeout]] = DEFAULT_TIMEOUT,
        custom_headers: Optional[Mapping[str, str]] = None,
    ):
        self._base_url = self._enforce_trailing_slash(URL(base_url))
        self._version = _version
        self.max_retries = max_retries
        self.timeout = timeout
        self._limits = limits
        self._custom_headers = custom_headers or {}
        self.backoff_base = 2

    def _enforce_trailing_slash(self, url: URL) -> URL:
        if url.raw_path.endswith(b"/"):
            return url
        return url.copy_with(raw_path=url.raw_path + b"/")

    def _build_headers(self, custom_headers) -> httpx.Headers:
        headers_dict = _merge_mappings(
            self.default_headers, self._custom_headers, custom_headers
        )
        headers = httpx.Headers(headers_dict)
        return headers or dict()

    def _build_params(self, custom_params) -> dict[str, str]:
        params = _merge_mappings(self.default_params, custom_params)
        return params

    @property
    def custom_auth(self) -> Union[httpx.Auth, None]:
        return None

    @property
    def auth_headers(self) -> dict[str, str]:
        return {}

    @property
    def default_headers(self) -> dict[str, str]:
        return {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }

    @property
    def default_params(self) -> dict[str, str]:
        return None

    def platform_headers(self) -> Dict[str, str]:
        return platform_headers(self._version)

    @retry(
        max_retries_func=lambda self: self.max_retries,
        exceptions=(httpx.RequestError,),
    )
    def _request(
        self, method: str, path: str, json=None, headers=None, **kwargs
    ) -> httpx.Response:
        url = urljoin(str(self._base_url), path)
        logger.debug(f"Requesting {method} {url}")
        merged_headers = self._build_headers(headers)
        merged_params = self._build_params(kwargs.get("params"))
        try:
            return self._client.request(
                method.upper(),
                url,
                json=json,
                headers=merged_headers,
                params=merged_params,
            )
        except httpx.TransportError as e:
            logger.error(f"Transport error: {e}.")
            raise e
        except httpx.RequestError as e:
            logger.error(f"Request failed with error: {e}.")
            raise e

    def get(self, path: str, headers=None, **kwargs) -> httpx.Response:
        """Make a GET request."""
        return self._request("GET", path, headers=headers, **kwargs)

    def post(self, path: str, json=None, headers=None, **kwargs) -> httpx.Response:
        """Make a POST request."""
        return self._request("POST", path, json=json, headers=headers, **kwargs)

    def patch(self, path: str, json=None, headers=None, **kwargs) -> httpx.Response:
        """Make a PATCH request."""
        return self._request("PATCH", path, json=json, headers=headers, **kwargs)

    def put(self, path: str, json=None, headers=None, **kwargs) -> httpx.Response:
        """Make a PUT request."""
        return self._request("PUT", path, json=json, headers=headers, **kwargs)

    def delete(self, path: str, headers=None, **kwargs) -> httpx.Response:
        """Make a DELETE request."""
        return self._request("DELETE", path, headers=headers, **kwargs)

    def close(self):
        """Close the underlying HTTPX client."""
        self.client.close()


class SyncAPIClient(BaseClient[httpx.Client]):
    _client: httpx.Client

    def __init__(
        self,
        base_url: Union[str, httpx.URL],
        max_retries: int = DEFAULT_MAX_RETRIES,
        timeout: Union[float, Timeout, None] = DEFAULT_TIMEOUT,
        limits: Limits = DEFAULT_CONNECTION_LIMITS,
        _version: Optional[str] = None,
        http_client: Optional[httpx.Client] = None,
        custom_headers: Optional[Mapping[str, str]] = None,
    ) -> None:

        if http_client is not None and not isinstance(
            http_client, httpx.Client
        ):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError(
                f"Invalid `http_client` argument; Expected an instance of `httpx.Client` but got {type(http_client)}"
            )

        super().__init__(
            _version=_version,
            limits=limits,
            # cast to a valid type because mypy doesn't understand our type narrowing
            timeout=cast(Timeout, timeout),
            base_url=base_url,
            max_retries=max_retries,
            custom_headers=custom_headers,
        )
        self._client = http_client or SyncHttpxClientWrapper(
            base_url=base_url,
            # cast to a valid type because mypy doesn't understand our type narrowing
            timeout=cast(Timeout, timeout),
            limits=limits,
            follow_redirects=True,
        )

    def is_closed(self) -> bool:
        return self._client.is_closed

    def close(self) -> None:
        """Close the underlying HTTPX client.

        The client will *not* be usable after this.
        """
        # If an error is thrown while constructing a client, self._client
        # may not be present
        if hasattr(self, "_client"):
            self._client.close()

    def _prepare_request(
        self,
        request: httpx.Request,  # noqa: ARG002
    ) -> None:
        """This method is used as a callback for mutating the `Request` object
        after it has been constructed.
        This is useful for cases where you want to add certain headers based off of
        the request properties, e.g. `url`, `method` etc.
        """
        return None


def _merge_mappings(*mappings):
    merged = {}
    for mapping in mappings:
        if mapping:
            merged.update(mapping)
    return merged


class AsyncAPIClient(BaseClient[httpx.AsyncClient]):
    pass


def get_platform() -> Platform:
    try:
        system = platform.system().lower()
        platform_name = platform.platform().lower()
    except Exception:
        return "Unknown"

    if "iphone" in platform_name or "ipad" in platform_name:
        # Tested using Python3IDE on an iPhone 11 and Pythonista on an iPad 7
        # system is Darwin and platform_name is a string like:
        # - Darwin-21.6.0-iPhone12,1-64bit
        # - Darwin-21.6.0-iPad7,11-64bit
        return "iOS"

    if system == "darwin":
        return "MacOS"

    if system == "windows":
        return "Windows"

    if "android" in platform_name:
        # Tested using Pydroid 3
        # system is Linux and platform_name is a string like 'Linux-5.10.81-android12-9-00001-geba40aecb3b7-ab8534902-aarch64-with-libc'
        return "Android"

    if system == "linux":
        # https://distro.readthedocs.io/en/latest/#distro.id
        distro_id = distro.id()
        if distro_id == "freebsd":
            return "FreeBSD"

        if distro_id == "openbsd":
            return "OpenBSD"

        return "Linux"

    return "Unknown"


def platform_headers(version: str) -> Dict[str, str]:
    return {
        "X-Stainless-Lang": "python",
        "X-Stainless-Package-Version": version,
        "X-Stainless-OS": str(get_platform()),
        "X-Stainless-Arch": str(get_architecture()),
        "X-Stainless-Runtime": get_python_runtime(),
        "X-Stainless-Runtime-Version": get_python_version(),
    }


def get_architecture() -> Arch:
    try:
        python_bitness, _ = platform.architecture()
        machine = platform.machine().lower()
    except Exception:
        return "unknown"

    if machine in ("arm64", "aarch64"):
        return "arm64"

    # TODO: untested
    if machine == "arm":
        return "arm"

    if machine == "x86_64":
        return "x64"

    # TODO: untested
    if python_bitness == "32bit":
        return "x32"

    return "unknown"


def get_python_runtime() -> str:
    try:
        return platform.python_implementation()
    except Exception:
        return "unknown"


def get_python_version() -> str:
    try:
        return platform.python_version()
    except Exception:
        return "unknown"


T = TypeVar('T', bound=Callable[..., httpx.Response])

# class APIResponseException(Exception):
#     def __init__(self, message: str, response: Optional[httpx.Response] = None):
#         self.message = message
#         self.response = response
#         self.code = response.status_code if response else None
        

class APIResponseManager(Generic[T]):
    api: Generic[T]
    response: httpx.Response = None

    def __init__(self, api: Generic[T]):
        self.api = api

    def __enter__(self) -> "APIResponseManager":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            logger.error("An error occurred in the API call.")
            if self.response:
                logger.error(f"Response status code: {self.response.status_code}")
                logger.error(f"Response content: {self.response.text}")
            logger.error(exc_val)
            traceback.print_exception(exc_type, exc_val, exc_tb)
            exit(1)
        return True

    def get_response(self, *args, **kwargs) -> httpx.Response:
        response: httpx.Response = self.api(*args, **kwargs)
        self.response = response
        return response




import logging

from bohrium_open_sdk.opensdk._resource import SyncAPIResource
from bohrium_open_sdk.opensdk._response import APIResponse
from bohrium_open_sdk.opensdk._base_client import APIResponseManager


logger = logging.getLogger(__name__)


class Sku(SyncAPIResource):

    def list(
        self,
        *,
        page: int = 1,
        page_size: int = 20,
        scene: str = "job",
        product_line: str = "bohrium",
        choose_type: str = "cpu",
    ):
        with APIResponseManager(self._client.get) as api:
            uri = f"/{self._client.api_prefix}/v1/open/sku/list"
            params = {
                "page": page,
                "pageSize": page_size,
            }
            scene and params.update({"scene": scene})
            product_line and params.update({"productLine": product_line})
            choose_type and params.update({"chooseType": choose_type})

            response = api.get_response(uri, params=params)
            return APIResponse(response).json
        
  


from pathlib import Path
from dataclasses import dataclass
from typing import Dict, List, Union, Optional

JobInputType = Union[int, str, float, bool, List, Dict, None]


@dataclass
class UploadInputItem:
    input_field: Optional[str]
    """_summary_
    InputItem remote planned path
    this value will auto be filled by the SDK
    """
    input_path: Optional[str]
    src: Union[str, Path]

    def __init__(self, *, src: Union[str, Path], input_field: Optional[str]=None):
        self.input_field = input_field
        self.src = src



import logging
from typing import Any, Dict, Mapping, Optional
from xml.dom import NotFoundErr

from httpx import Response

from . import _exceptions
from ._exceptions import APIStatusError

logger = logging.getLogger(__name__)


class APIErrorResponse:
    def __init__(self, message: str, status_code: int):
        self.message = message
        self.status_code = status_code

    def to_dict(self) -> Dict[str, Any]:
        return {"error": self.message, "status_code": self.status_code}

    def __repr__(self) -> str:
        return f"<APIErrorResponse(status_code={self.status_code}, message='{self.message}')>"


class APIResponse:
    def __init__(self, response: Response, request: Any = None):
        self._response = response
        self._request = request
        self.status_code = response.status_code
        self.headers = response.headers
        self.raise_for_status()
        self._json = self._parse_json(response)
        self._data = self._fetch_data(response)

    def _parse_json(self, response: Response) -> Optional[Any]:
        """Parses the JSON content of the response."""
        try:
            return response.json()
        except ValueError:
            return APIErrorResponse("Invalid JSON response", response.status_code)

    def _fetch_data(self, response: Response) -> Optional[Any]:
        try:
            return response.json().get("data")
        except ValueError:
            return APIErrorResponse("Invalid JSON response", response.status_code)

    @property
    def data(self) -> Optional[Any]:
        try:
            return DotDict(self._json["data"])
        except KeyError:
            raise NotFoundErr("No data found in response")

    @property
    def json(self) -> Optional[Any]:
        return self._json

    def is_success(self) -> bool:
        """Returns True if the response status code indicates success."""
        return 200 <= self.status_code < 300

    def raise_for_status(self):
        """Raises an HTTPError if the response status code indicates an error."""
        if not self.is_success():
            err_msg = f"HTTP Error {self.status_code} for url {self._response.url}, response: {self._response.text}"
            logging.error(err_msg)
            err = self._make_status_error(
                err_msg,
                body=None,
                response=self._response,
            )
            raise err

    def __repr__(self) -> str:
        return (
            f"<APIResponse(status_code={self.status_code}, "
            f"headers={dict(self.headers)}, "
            f"json={self._json})>"
        )

    def _make_status_error(
        self,
        err_msg: str,
        *,
        body: object,
        response: Response,
    ) -> APIStatusError:
        data = body.get("error", body) if isinstance(body, Mapping) else body
        if response.status_code == 400:
            return _exceptions.BadRequestError(err_msg, response=response, body=data)

        if response.status_code == 401:
            return _exceptions.AuthenticationError(
                err_msg, response=response, body=data
            )

        if response.status_code == 403:
            return _exceptions.PermissionDeniedError(
                err_msg, response=response, body=data
            )

        if response.status_code == 404:
            return _exceptions.NotFoundError(err_msg, response=response, body=data)

        if response.status_code == 409:
            return _exceptions.ConflictError(err_msg, response=response, body=data)

        if response.status_code == 422:
            return _exceptions.UnprocessableEntityError(
                err_msg, response=response, body=data
            )

        if response.status_code == 429:
            return _exceptions.RateLimitError(err_msg, response=response, body=data)

        if response.status_code >= 500:
            return _exceptions.InternalServerError(
                err_msg, response=response, body=data
            )
        return APIStatusError(err_msg, response=response, body=data)


class DotDict(dict):
    """A dictionary that supports dot notation for accessing values."""

    def __getattr__(self, attr):
        try:
            value = self[attr]
            if isinstance(value, dict):
                return DotDict(value)
            return value
        except KeyError:
            raise AttributeError(f"No such attribute: {attr}")

    def __setattr__(self, attr, value):
        self[attr] = value

    def __delattr__(self, attr):
        try:
            del self[attr]
        except KeyError:
            raise AttributeError(f"No such attribute: {attr}")



from typing import Any, Callable, Dict, Optional


def dict_traverser(
    data: Any,
    handlers: Dict[Any, Callable] = {},
    key: Optional[str] = None,
):
    for type, handler in handlers.items():
        if isinstance(data, type):
            new_res = handler(key, data)
            return new_res

    if isinstance(data, list) or isinstance(data, tuple):
        if isinstance(data, list):
            return [dict_traverser(item, handlers, key) for item in data]
        return (dict_traverser(item, handlers, key) for item in data)
    if not isinstance(data, dict):
        return data

    new_res = {}
    for k, v in data.items():
        new_res[k] = dict_traverser(v, handlers, k)
    return new_res



import logging

from bohrium_open_sdk.opensdk._resource import SyncAPIResource
from bohrium_open_sdk.opensdk._response import APIResponse
from bohrium_open_sdk.opensdk._base_client import APIResponseManager


log = logging.getLogger(__name__)


class User(SyncAPIResource):

    def get_info(self):
        response = self._client.get("openapi/v1/ak/user")
        return APIResponse(response).json

    def list_project(self):
        with APIResponseManager(self._client.get) as api:
            uri = f"/{self._client.api_prefix}/v1/open/user/project/list"

            response = api.get_response(uri)
            return APIResponse(response).json
        
    